from silmused.tests.TestDefinition import TestDefinition


# TODO raname to TableDataTest ??
class DataTest(TestDefinition):
    def __init__(self, name, title=None, column_name=None, should_exist=True, where=None, join=None, description=None,
                 expected_value=None, points=0):

        if column_name is not None and not isinstance(column_name, str):
            raise Exception('Parameter "column_name" must be a string')

        super().__init__(
            name=name,
            title=title,
            where=where,
            join=join,
            points=points,
            description=description,
            query=f"SELECT {column_name if column_name is not None else '*'} FROM {name}",
            should_exist=should_exist,
            expected_value=expected_value,
        )

        self.column_name = column_name
        self.where = where
        self.join = join

    def execute(self, cursor):
        cursor.execute(self.query)
        result = cursor.fetchall()

        if self.expected_value is None:
            if self.should_exist:
                return super().response(
                    len(result) > 0,
                    f"Correct, results found for table {self.name} and column(s) {self.column_name}",
                    f"Expected to find results for table {self.name} and column(s) {self.column_name} but none were found",
                )
            else:
                return super().response(
                    len(result) == 0,
                    f"Correct, no results found for table {self.name} and column(s) {self.column_name} ",
                    f"Expected to find nor results for table {self.name} and column(s) {self.column_name} but some were found",
                )
        else:
            if self.should_exist:
                # TODO add type check

                return super().response(
                    str(result[0][0]) == str(self.expected_value),
                    f"Correct value found for table {self.name} and column(s) {self.column_name}",
                    f"Expected to find {self.expected_value} for table {self.name} and column(s) {self.column_name} but found {result[0][0]}",
                )
            else:
                return super().response(
                    str(result[0][0]) != str(self.expected_value),
                    f"Correct, {self.expected_value} does not equal {result[0][0]} in  table {self.name} and column(s) {self.column_name} ",
                    f"Expected {self.expected_value} to not equal {result[0][0]} in table {self.name} and column(s) {self.column_name}",
                )



from silmused.tests.TestDefinition import TestDefinition
from silmused.utils import list_to_string


class FunctionTest(TestDefinition):
    def __init__(self, name, title=None, arguments=None, column_name=None, where=None, description=None, expected_value=None, expected_count=None, number_of_parameters=None, points=0):
        super().__init__(
            name=name,
            title=title,
            where=where,
            points=points,
            arguments=arguments,
            description=description,
            expected_value=expected_value,
            expected_count=expected_count,
            query=f"SELECT {column_name if column_name is not None else '*'} FROM {name}({list_to_string(arguments if arguments is not None else '')})"  # TODO implement parameters,
        )

        self.number_of_parameters = number_of_parameters

    def execute(self, cursor):
        # TODO small tests in separate functions

        # Check that function name exists
        cursor.execute(f"SELECT * FROM pg_catalog.pg_proc WHERE proname='{self.name}'")
        if len(cursor.fetchall()) <= 0:
            return super().response(
                False,
                ''
                f"Function {self.name} was not found"
            )

        # Check that function number of arguments is correct
        if self.number_of_parameters is not None:
            cursor.execute(f"SELECT pronargs FROM pg_catalog.pg_proc WHERE proname='{self.number_of_parameters}'")
            number_of_parameters_result = cursor.fetchall()[0][0]

            if not number_of_parameters_result == self.number_of_parameters:
                return super().response(
                    False,
                    ''
                    f"Expected {self.number_of_parameters} for function {self.name} but found {number_of_parameters_result}"
                )
        # Check that type is correct
        cursor.execute(f"SELECT routine_name FROM information_schema.routines WHERE routine_type = 'FUNCTION' AND routine_name='{self.name}'")
        if not len(cursor.fetchall()) > 0:
            return super().response(
                False,
                ''
                f"Expected function {self.name} to be of type FUNCTION"
            )

        cursor.execute(self.query)
        result = cursor.fetchall()

        if self.expected_value is None:
            if self.expected_count is None:
                return super().response(
                    len(result) > 0,
                    f"Correct count > 0 for function {self.name}({list_to_string(self.arguments)})",
                    f"Expected count > 0 for function {self.name}({list_to_string(self.arguments)}) but none was found",
                )
            else:
                return super().response(
                    len(result) == self.expected_count,
                    f"The count is correct for function {self.name}({list_to_string(self.arguments)}) -> {self.expected_count}",
                    f"Expected count {self.expected_count} for function {self.name}({list_to_string(self.arguments)}) but got {len(result)}",
                )
        else:
            # TODO when type is decimal but we give float, we get an error
            # if type(result[0][0]) != type(self.expected_value):
            #     return super().response(
            #         False,
            #         'Correct',
            #         f"Expected type {type(self.expected_value)} but got {type(result[0][0])}",
            #     )
            return super().response(
                str(result[0][0]) == str(self.expected_value),
                'Correct',
                f"Expected {self.expected_value} but got {result[0][0]}",
            )



from silmused.tests.TestDefinition import TestDefinition
from silmused.utils import list_to_string


class ProcedureTest(TestDefinition):
    def __init__(self, name, arguments, title=None, number_of_columns=None, description=None, expected_value=None, expected_count=None, pre_query=None, after_query=None, points=0):
        if after_query is None:
            raise Exception('Parameter "after_query" is required')

        super().__init__(
            name=name,
            title=title,
            points=points,
            arguments=arguments,
            description=description,
            expected_value=expected_value,
            expected_count=expected_count,
            pre_query=pre_query,
            query=f"CALL {name}({list_to_string(arguments)})",
            after_query=after_query,
        )

        self.number_of_columns = number_of_columns

    def execute(self, cursor):
        if self.number_of_columns is not None:
            cursor.execute(f"SELECT pronargs FROM pg_catalog.pg_proc WHERE proname = '{self.name}'")
            if not cursor.fetchall()[0][0] == self.number_of_columns:
                return False, f"Procedure's \"{self.name}\" number of parameters is wrong",

        if self.pre_query is not None:
            cursor.execute(self.pre_query)

        cursor.execute(self.query)

        cursor.execute(self.after_query)

        result = cursor.fetchall()

        return super().response(
            len(result) > 0,
            f"Correct count > 0 for procedure \"{self.name}({list_to_string(self.arguments)})\"",
            f"Expected count > 0 for procedure \"{self.name}({list_to_string(self.arguments)})\" but none was found",
        )


import json
import os
import subprocess
import sys

import psycopg2
from uuid import uuid4
from datetime import datetime
from silmused.version import __version__

from psycopg2.extensions import ISOLATION_LEVEL_AUTOCOMMIT


def _results_to_string(results):
    string = ''
    for result in results:
        if result.get('type') != 'execution' and result.get('message') is not None:
            string += str(result.get('message')) + '\n'

    print(string)


class Runner:
    def __init__(self, backup_file_path, tests, lang='en', test_name='', db_user='postgres', db_host='localhost', db_password='postgresql', db_port='5432'):
        self.file_path = backup_file_path
        self.tests = tests
        self.test_name = test_name
        self.db_user = db_user
        self.db_name = ''
        self.db_password = db_password
        self.db_host = db_host
        self.db_port = db_port
        self.lang = lang

        self.db_name = f"db{'_' + self.test_name if self.test_name != '' else ''}_{self.file_path.split('/')[-1].split('.')[0]}_{str(uuid4()).replace('-', '_')}"

        self.results = []

        if self._file_is_valid_pg_dump():
            self._create_db_from_psql_dump()
            self.results = self._run_tests()
            # _results_to_string(self.results)
        elif self._file_is_valid_pg_insert():
            self._create_db_from_psql_insert()
            self.results = self._run_tests()
            # _results_to_string(self.results)
        else:
            print('Error: File is not a valid PostgresSql dump or insert file!')

    def _file_is_valid_pg_dump(self):
        if not os.path.isfile(self.file_path):
            return False

        try:
            with open(self.file_path, 'rb') as file:
                magic_number = file.read(5)
                return magic_number == b'PGDMP'
        except IOError:
            return False

    def _create_db_from_psql_dump(self):
        subprocess.run(["createdb", "-U", self.db_user, self.db_name])

        subprocess.run(
            ["psql", "-U", self.db_user, "-d", self.db_name, "-c", "DROP SCHEMA public CASCADE;"],
            stdout=subprocess.DEVNULL,
            stderr=subprocess.DEVNULL
        )

        subprocess.run(
            ["pg_restore", "-U", self.db_user, "-d", self.db_name, "--no-owner", "--no-acl", "-1", self.file_path]
        )

    def _file_is_valid_pg_insert(self):
        if not os.path.isfile(self.file_path):
            return False

        if not self.file_path.lower().endswith('.sql'):
            return False

        try:
            with open(self.file_path, 'r') as file:
                lines = file.readlines()
                return any(line.strip().startswith("INSERT") for line in lines)
        except IOError:
            return False

    def _create_db_from_psql_insert(self):
        connection = self._connect(db_name='postgres')
        connection.set_isolation_level(ISOLATION_LEVEL_AUTOCOMMIT)
        cursor = connection.cursor()

        try:
            cursor.execute(f"CREATE DATABASE {self.db_name};")
        except Exception as exception:
            print(f"CREATE DATABASE failed: {exception}")
        finally:
            cursor.close()
            connection.close()

        connection = self._connect()
        cursor = connection.cursor()

        try:
            with open(self.file_path, 'r') as file:
                sql_script = file.read()
            #if self.db_user != 'silmus':
            cursor.execute('DROP SCHEMA IF EXISTS public')
            cursor.execute(sql_script)
            connection.commit()
        except Exception as exception:
            print(f"Sql INSERT import failed: {exception}")
        finally:
            cursor.close()
            connection.close()

    def _run_tests(self):
        connection = self._connect()
        cursor = connection.cursor()

        try:
            results = []

            for test in self.tests:
                results.append(test.run(cursor))

            return results
        except Exception as exception:
            print(f"Sql TEST RUN failed: {exception}")
        finally:
            cursor.close()
            connection.close()

    def _connect(self, db_name=None):
        connection_layer = psycopg2.connect(
            host=self.db_host,
            port=self.db_port,
            database=self.db_name if db_name is None else db_name,
            user=self.db_user,
            password=self.db_password
        )

        return connection_layer

    def _checks_to_object(self, checks):
        outputs = []
        output_pass = True

        # TODO should be recursive
        points_max = 0
        points_actual = 0

        for check in checks:
            if check.get('type') == 'execution':
                continue
            elif check.get('type') == 'message':
                outputs.append({
                    "title": str(check.get('message')),
                    "status": 'PASS'
                })
                continue

            points = check.get('points') if check.get('points') is not None else 0
            points_max += points
            points_actual += points if check.get('is_success') else 0

            output = {}
            output["title"] = check.get('title')
            output["status"] = 'PASS' if check.get('is_success') else 'FAIL'
            output_pass = True if check.get('is_success') and output_pass else False
            output["feedback"] = str(check.get('message')) if not check.get('is_success') else ''

            outputs.append(output)

        return points_max, points_actual, outputs, output_pass

    def _results_to_object(self):
        tests = []

        points_max = 0
        points_actual = 0

        for result in self.results:
            if result.get('type') == 'execution':
                continue
            elif result.get('type') == 'message':
                tests.append({
                    "title": str(result.get('message')),
                    "status": 'PASS'
                })
                continue

            output = {}

            if result.get('type') == 'checks_layer':
                checks_points_max, checks_points_actual, checks_outputs, output_pass = self._checks_to_object(result.get('checks'))
                points_max += checks_points_max
                points_actual += checks_points_actual
                output["checks"] = checks_outputs
                output["status"] = 'PASS' if output_pass else 'FAIL'
            else:
                points = result.get('points') if result.get('points') is not None else 0
                points_max += points
                points_actual += points if result.get('is_success') else 0
                output["status"] = 'PASS' if result.get('is_success') else 'FAIL'

            
            output["title"] = result.get('title')
            #print(result.get('message'))
            if result.get('message') is not None:
                output["exception_message"] = str(result.get('message'))

            tests.append(output)
        return tests, points_max, points_actual

    def get_results(self):
        try:
            tests, points_max, points_actual = self._results_to_object()
            praks = True if len(tests) > 0 and points_max == 0 and points_actual == 0 else False
            if praks:
                points_max = 1
                points_actual = 1

            output = {
                "result_type": "OK_V3",
                "points": round(100 * points_actual / points_max),
                "producer": f"silmused {__version__}",
                "finished_at": datetime.utcnow().strftime("%Y-%m-%dT%H:%M:%SZ"),
                "tests": tests
            }
            return json.dumps(output, ensure_ascii=False)
        except:
            print(sys.exc_info())
            return json.dumps({
              "result_type": "OK_V3",
              "producer": f"silmused {__version__}",
              "finished_at": datetime.utcnow().strftime("%Y-%m-%dT%H:%M:%SZ"),
              "points": 0,
            }, ensure_ascii=False)


import sys


# TODO this should not be usable without a child
class TestDefinition():
    def __init__(self, name, points, title='', where=None, join=None, should_exist=True, query='', description=None,
                 arguments=None, expected_value=None, expected_count=None, pre_query=None, after_query=None):
        if arguments is not None and not isinstance(arguments, list):
            raise Exception('Parameter "arguments" must be a list')

        if expected_count is not None and not isinstance(expected_count, int):
            raise Exception('Parameter "expected_count" must be an integer')

        if not isinstance(points, int) and not isinstance(points, float):
            raise Exception('Parameter "points" must be either an integer or a float')

        if expected_value is not None and expected_count is not None:
            raise Exception('Both expected_value and check_count cannot be specified in a single test')

        query_builder = query

        # TODO right now a single join is possible (without a hack)
        if join is not None:
            query_builder += f" JOIN {join}"

        if where is not None:
            query_builder += f" WHERE ({where})"

        self.title = title
        self.points = points
        self.name = name
        self.description = description
        self.arguments = arguments # TODO arguments could be a class
        self.expected_value = expected_value
        self.expected_count = expected_count
        self.query = query_builder
        self.pre_query = pre_query
        self.after_query = after_query
        self.should_exist = should_exist  # TODO should be renamed to something more descriptive (should_be_false/falsy)

    # TODO should be callable only inside the scope
    def execute(self, cursor):
        raise NotImplementedError('Method "execute" not implemented')

    def run(self, cursor):
        try:
            # TODO could executing of pre and/or after queries be handled here?
            return self.execute(cursor)
        except:
            # TODO better handler for rollback?
            # TODO better error message?

            cursor.execute('ROLLBACK')
            return self.response(
                False,
                message_failure=sys.exc_info()
            )

    # TODO should be callable only inside the scope
    def response(self, is_success, message_success=None, message_failure=None, points=None):
        if is_success:
            message_statement = 'Correct' if message_success is None else message_success
        else:
            message_statement = 'Wrong' if message_failure is None else message_failure

        return {
            'is_success': is_success,
            'message': message_statement,
            'points': points if points is not None else self.points,
            'description': self.description,
            'query': self.query,
            'pre_query': self.pre_query,
            'after_query': self.after_query,
            'should_exist': self.should_exist,
            'title': self.title,
        }



import sys

from silmused.Runner import Runner

'''
CLI to integrate the solution with e.g. a Docker sultion.

NB! The tests file must include an array with the key "tests"!!
'''


def main():
    sql_file_path = sys.argv[1]
    tests_path = sys.argv[2]
    db_user = sys.argv[3] if len(sys.argv) > 3 else 'postgres'
    db_host = sys.argv[4] if len(sys.argv) > 4 else 'localhost'
    db_port = sys.argv[5] if len(sys.argv) > 5 else '5432'
    db_password = sys.argv[6] if len(sys.argv) > 6 else 'postgresql'
    
    with open(tests_path, 'r') as file:
        tests_file_content = file.read()

    local_vars = {}
    exec(tests_file_content, globals(), local_vars)
    tests = local_vars.get('tests')

    r=Runner(sql_file_path, tests, db_user=db_user, db_host=db_host, db_port=db_port, db_password=db_password)
    print(r.get_results())


from silmused.tests.TestDefinition import TestDefinition
from silmused.utils import list_to_string


class StructureTest(TestDefinition):
    def __init__(self, name, title=None, column_name=None, arguments=None, expected_value=None, should_exist=True, where=None, description=None, points=0):
        if column_name is not None and not isinstance(column_name, str):
            raise Exception('Parameter "column_name" must be a string')

        query = f"SELECT {list_to_string(arguments)[1:-1] if arguments is not None else '*'} FROM information_schema.columns WHERE table_name = '{name}'"

        if column_name is not None:
            query += f" AND column_name = '{column_name}'"

        super().__init__(
            name=name,
            title=title,
            where=where,
            points=points,
            arguments=arguments,
            description=description,
            expected_value=expected_value,
            query=query,
            should_exist=should_exist,
        )

        self.column_name = column_name
        self.where = where

    def execute(self, cursor):
        # TODO feedback is really bad
        cursor.execute(self.query)
        result = cursor.fetchall()

        if self.expected_value is not None:
            if self.should_exist:
                return super().response(
                    len(result) != 0 and result[0][0] == self.expected_value,
                    f"Correct, expected {self.query}", # TODO enhance feedback
                    f"Wrong, did not expect {self.query}", # TODO enhance feedback
                )
            else:
                return super().response(
                    len(result) == 0 or result[0][0] != self.expected_value,
                    f"Correct did not want {self.query}", # TODO enhance feedback
                    f"Wrong this should not exist {self.query}", # TODO enhance feedback
                )
        else:
            if self.should_exist:
                return super().response(
                    len(result) > 0,
                    f"Correct, column or table {self.column_name if self.column_name is not None else self.name} found in table {self.name}",
                    f"Expected to find column or table {self.column_name if self.column_name is not None else self.name} but none were found in table {self.name}",
                )
            else:
                return super().response(
                    len(result) == 0,
                    f"Correct no column or table named {self.column_name if self.column_name is not None else self.name} found in table {self.name}",
                    f"Expected to not find column or table {self.column_name if self.column_name is not None else self.name} in table {self.name}",
                )



from silmused.tests.TestDefinition import TestDefinition


class TriggerTest(TestDefinition):
    def __init__(self, name, title=None, description=None, points=0, arguments=None, action_timing=None):
        super().__init__(
            name=name,
            title=title,
            points=points,
            description=description,
            arguments=arguments,
            query=f"SELECT * FROM information_schema.views WHERE table_name = '{name}'",
        )

        self.action_timing = action_timing

    def execute(self, cursor):
        # TODO small tests in separate functions
        cursor.execute(f"SELECT trigger_name FROM information_schema.triggers WHERE trigger_name = '{self.name}'")

        if len(cursor.fetchall()) <= 0:
            return super().response(
                False,
                ''
                f"Trigger {self.name} was not found",
            )

        errors = []

        if len(self.arguments) > 0:
            for manipulation in self.arguments:
                cursor.execute(
                    f"SELECT trigger_name FROM information_schema.triggers WHERE trigger_name = '{self.name}' AND event_manipulation = '{manipulation}'")
                if len(cursor.fetchall()) <= 0:
                    errors.append(f"manipulation {manipulation} was not found")

        # TODO this is a bit of a copy-paste
        cursor.execute(f"SELECT trigger_name FROM information_schema.triggers WHERE trigger_name = '{self.name}' AND action_timing = '{self.action_timing}'")
        if len(cursor.fetchall()) <= 0:
            errors.append(f"action timing {self.action_timing} was not found")

        return super().response(
            len(errors) == 0,
            f"Correct, trigger {self.name} definition is correct",
            f"Trigger {self.name} had the following errors: {', '.join(errors)}",
        )



from silmused.tests.TestDefinition import TestDefinition


class ConstraintTest(TestDefinition):
    def __init__(self, name, title=None, column_name=None, constraint_name=None, constraint_type=None, description=None, should_exist=True, points=0):
        query = f"SELECT * FROM information_schema.table_constraints WHERE table_name = '{name}'"

        if column_name is not None:
            query += f" AND constraint_name LIKE '%{column_name}%'"
        if constraint_name is not None:
            query += f" AND constraint_name = '{constraint_name}'"
        if constraint_type is not None:
            query += f" AND constraint_type = '{constraint_type}'"

        super().__init__(
            name=name,
            title=title,
            points=points,
            description=description,
            query=query,
            should_exist=should_exist,
        )

        self.constraint_name = constraint_name
        self.constraint_type = constraint_type
        self.column_name = column_name

        self.feedback = f"table {self.name}"

        if self.column_name is not None:
            self.feedback += f" with column = '{self.column_name}'"
        if self.constraint_name is not None:
            self.feedback += f" with constraint name = '{self.constraint_name}'"
        if self.constraint_type is not None:
            if self.constraint_name is None:
                self.feedback += f" with constraint type = '{self.constraint_type}'"
            else:
                self.feedback += f" and type = '{self.constraint_type}'"

    def execute(self, cursor):
        cursor.execute(self.query)
        result = cursor.fetchall()

        if self.should_exist:
            return super().response(
                len(result) > 0,
                f"Correct, {self.feedback} was found",
                f"Expected to find {self.feedback} but none were found",
            )
        else:
            return super().response(
                len(result) == 0,
                f"Correct, {self.feedback} was not found",
                f"Expected to no find {self.feedback}",
            )


from uuid import uuid4


class ChecksLayer:
    def __init__(self, title=f"test_{str(uuid4())}", tests=[]):
        self.title = title
        self.tests = tests

    def run(self, cursor):
        results = []
        for test in self.tests:
            results.append(test.run(cursor))

        return {
            'title': self.title,
            'type': 'checks_layer',
            'checks': results,
        }



from silmused.tests.TestDefinition import TestDefinition


class IndexTest(TestDefinition):
    def __init__(self, name, title=None, description=None, points=0):
        super().__init__(
            name=name,
            title=title,
            points=points,
            description=description,
            query=f"SELECT * FROM pg_indexes WHERE indexname = '{name}'",
        )

    def execute(self, cursor):
        cursor.execute(self.query)
        result = cursor.fetchall()

        return super().response(
            len(result) > 0,
            f"Correct, index was {self.name} found",
            f"Expected to find index {self.name} but none were found",
        )


from setuptools import setup, find_namespace_packages
exec(open('silmused/version.py').read())

setup(
    name='silmused',
    version=__version__,
    license='MIT License',
    author='Karel Paan, Martti Kakk',
    author_email='paan.karel@gmail.com',
    packages=find_namespace_packages(),
    url='https://github.com/kare22/silmused',
    install_requires=[
        'psycopg2'
    ],
    entry_points={
        'console_scripts': [
            'silmused=silmused.cli:main',
        ],
    },
)


from silmused.tests.TestDefinition import TestDefinition


# TODO Right now it is only possible to check the existence of a view
# Probably there is no need for this as it is covered by table tests

# We can use it to test the existence of materialized views
class ViewTest(TestDefinition):
    def __init__(self, name, title=None, description=None, points=0):
        super().__init__(
            name=name,
            title=title,
            points=points,
            description=description,
            query=f"SELECT * FROM information_schema.views WHERE table_name = '{name}'",
        )

    def execute(self, cursor):
        cursor.execute(self.query)
        result = cursor.fetchall()

        return super().response(
            len(result) > 0,
            f"Correct, view {self.name} found",
            f"Expected to find view {self.name} but none were found",
        )